# this is the first working image processing script for detecting signs. Used on Axels indoor robot. 
# Info about the robot here: http://axelsdiy.brinkeby.se/?page_id=1514

# Axel Brinkeby 2018-01-17

# import the necessary packages
from picamera.array import PiRGBArray
from picamera import PiCamera
import RPi.GPIO as GPIO
import time
import serial
import cv2
import timeit

import numpy as np

def order_points(pts):
	# initialize a list of coordinates that will be ordered
	# such that the first entry in the list is the top-left,
	# the second entry is the top-right, the third is the
	# bottom-right, and the fourth is the bottom-left
	rect = np.zeros((4, 2), dtype="float32")
 
	# the top-left point will have the smallest sum, whereas
	# the bottom-right point will have the largest sum
	s = pts.sum(axis=1)
	rect[0] = pts[np.argmin(s)]
	rect[2] = pts[np.argmax(s)]
 
	# now, compute the difference between the points, the
	# top-right point will have the smallest difference,
	# whereas the bottom-left will have the largest difference
	diff = np.diff(pts, axis=1)
	rect[1] = pts[np.argmin(diff)]
	rect[3] = pts[np.argmax(diff)]
 
	# return the ordered coordinates
	return rect
	
leds = [5, 6, 13, 19, 26]

def init_LEDs():
	GPIO.setmode(GPIO.BCM)
	for i in range(len(leds)):
		GPIO.setup(leds[i],GPIO.OUT)
		GPIO.output(leds[i],GPIO.LOW)
		print('LED number {} on GPIO {} initialized'.format(i, leds[i]))
		
def init_Serial():
	serialport = serial.Serial(
		port = '/dev/ttyAMA0',
		baudrate = 115200,
		parity = serial.PARITY_NONE,
		stopbits = serial.STOPBITS_ONE,
		bytesize = serial.EIGHTBITS,
		timeout = 1
	)
	print "Serial port opend sucessfully."
	return serialport

def toggle_hartbeat_LED():
	global hartbeatLEDstate; 
	if (hartbeatLEDstate):
		GPIO.output(leds[0],GPIO.HIGH)
		hartbeatLEDstate = False
	else:
		GPIO.output(leds[0],GPIO.LOW)
		hartbeatLEDstate = True
		
fileNames = ["_turnAround.png", "_park.png", "_left.png", "_right.png", "_spinAround.png", "_charge.png"]
symbolImages = []

print "Program started" 

hartbeatLEDstate = False
sign = 0 	# index of the sign currently visible, 0 = none, -1 = somthing unknown, 1 = turnaround, 2 = park
direction = 0	# the diretion to the sign
distance = 0  	# distance in cm from the sign from the camera
timeout = 10*60 #5*60		# stop program after this many seconds, set timeout = 0 to never stop

init_LEDs()
serialport = init_Serial()
 
imageSizeWidth = 1280
imageSizeHeight = 720
camera = PiCamera()
camera.resolution = (imageSizeWidth, imageSizeHeight)
camera.framerate = 30
rawCapture = PiRGBArray(camera, size=(imageSizeWidth, imageSizeHeight))

# allow the camera to warmup
time.sleep(0.1)
print "Camera started sucessfully." 

for i in range(len(fileNames)):
	temp = cv2.imread(fileNames[i])
	temp2 = cv2.cvtColor(temp, cv2.COLOR_RGB2GRAY)
	symbolImages.append(temp2)

print "Number of images loaded: "
print len(symbolImages)

#temp = cv2.imread("_turnAround.png")
#turnAroundImage = cv2.cvtColor(temp, cv2.COLOR_BGR2GRAY)
#temp = cv2.imread("_park.png")
#parkImage = cv2.cvtColor(temp, cv2.COLOR_BGR2GRAY)

startTime = timeit.default_timer(); 
 
# capture frames from the camera
for frame in camera.capture_continuous(rawCapture, format="bgr", use_video_port=True):
	
	sign = 0
	
	# grab the raw NumPy array representing the image
	image = frame.array
	#cv2.imwrite('original.jpg',image)				# save original image
	
	gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)	
	blur = cv2.blur(gray,(5,5))	
	edges = cv2.Canny(blur,10,80)	
	im2, contours, hierarchy = cv2.findContours(edges,cv2.RETR_TREE,cv2.CHAIN_APPROX_SIMPLE)
	
	#cv2.imwrite('edges.jpg',im2)					# save edge detecting image
	
	foundMatch = False
	
	for contour in contours:
		approx = cv2.approxPolyDP(contour,0.01*cv2.arcLength(contour,True),True)
		if len(approx) == 4:
			area = cv2.contourArea(approx)
			if area > 4000:		
				#print "Found something, processing..."
				cv2.drawContours(image, approx, -1, (0,0,255), 5)		# color is GBR standard, not RGB
				#cv2.imwrite('approx.jpg',image) 						# save image with marked area
				
				#points = np.array(approx, dtype = "int")
				points = np.array([[approx[0][0][0], approx[0][0][1]], [approx[1][0][0], approx[1][0][1]], [approx[2][0][0], approx[2][0][1]], [approx[3][0][0], approx[3][0][1]]], dtype = "float32")
				
				#print "Corner points in the image (after sorting):"
				points = order_points(points)
				#print(points)
				
				#print "Corner points in the image:"				
				points2 = np.array([[0,0],[255,0],[255,255], [0,255]], dtype = "float32")
				#print(points2)	
				
				direction = (imageSizeWidth / 2) - ((points[0][0] + points[1][0] + points[2][0] + points[3][0]) / 4	)		
				markerSize = points[1][0] - points[0][0]
				if markerSize == 0:
					distance = 0
				else:
					distance = (7 * 550) / markerSize
				
				M = cv2.getPerspectiveTransform(points,points2)
				
				
				warped = cv2.warpPerspective(image,M,(256,256))
				#cv2.imwrite('warped.jpg',warped)						# save the dewarped image
				#warped = cv2.cvtColor(warped, cv2.COLOR_BGR2GRAY)
				
				#(height, width, channels) = warped.shape
				#print "Image info: "
				#print height, width, channels
								
				if (len(warped.shape) == 2):
					break

				warpedGrayscale = cv2.cvtColor(warped, cv2.COLOR_BGR2GRAY)	
				
				(thresh, warpedBinImage) = cv2.threshold(warpedGrayscale, 128, 255, cv2.THRESH_BINARY | cv2.THRESH_OTSU)	
				
				#cv2.imwrite('warped_threshold.jpg',warpedBinImage)			# save the dewarped binary threshold image
				
				sign = -1
				
				#diffImg = symbolImages[0] - warpedBinImage	    # save some examples of the differance images
				#diffTurnAround = cv2.countNonZero(diffImg)			
				#cv2.imwrite('diffImg0.jpg',diffImg)	

				#diffImg = symbolImages[1] - warpedBinImage	
				#diffPark = cv2.countNonZero(diffImg)						
				#cv2.imwrite('diffImg1.jpg',diffImg)			
				
				for idx, image in enumerate(symbolImages):
					diffImg = image - warpedBinImage	
					difference = cv2.countNonZero(diffImg)
					#print difference
					
					if (difference < 8000):
						sign = idx + 1
						foundMatch = True
						
				if (foundMatch):
					break
									

	if sign == 0: 
		direction = 0
		distance = 0
		GPIO.output(leds[1],GPIO.LOW)
		GPIO.output(leds[2],GPIO.LOW)
		
	if sign == 1:
		GPIO.output(leds[1],GPIO.HIGH)
		GPIO.output(leds[2],GPIO.LOW)
	
	if sign == 2:
		GPIO.output(leds[1],GPIO.LOW)
		GPIO.output(leds[2],GPIO.HIGH)
		
	print('Index: {}   Distance: {}   Direction: {}'.format(sign, distance, direction))
	
	serialport.write(('%d,%d,%d\n'%(sign, distance, direction)).encode('utf-8'))
	
	# clear the stream in preparation for the next frame
	rawCapture.truncate(0)
	
	toggle_hartbeat_LED()	
	currentTime = timeit.default_timer(); 	
	if (timeout != 0 and currentTime > (startTime + timeout)):	# stop program after some time
		break
		
# make sure all the LEDs are off when the program exits
for i in range(len(leds)):		
	GPIO.output(leds[i],GPIO.LOW)